@@ -1,11 +1,12 @@
 """Main module."""
 import json
 import logging
+import click
 import os
 import sys
 from datetime import datetime
 from typing import List
-
+from distutils.util import strtobool
 
 import geofeather as gf
 import geopandas as gpd
@@ -197,7 +198,7 @@ def geocode(
     else:
         logging.info("GADM has not been provided; loading now.")
 
-        if admin == 'admin0':
+        if admin in ['admin0','country']:
             gadm_fn = f"gadm36_2.feather"
             gadmDir = f"{download_data_folder}/{gadm_fn}"
             gadm = gf.from_geofeather(gadmDir)
@@ -209,30 +210,33 @@ def geocode(
             gadmDir = f"{download_data_folder}/{gadm_fn}"
             gadm = gf.from_geofeather(gadmDir)
             gadm["country"] = gadm["NAME_0"]
-            gadm["state"] = gadm["NAME_1"]
+            #gadm["state"] = gadm["NAME_1"]
             gadm["admin1"] = gadm["NAME_1"]
-            gadm = gadm[["geometry", "country", "state", "admin1"]]
+            #gadm = gadm[["geometry", "country", "state", "admin1"]]
+            gadm = gadm[["geometry", "country", "admin1"]]
 
         elif admin == "admin2":
             gadm_fn = f"gadm36_2.feather"
             gadmDir = f"{download_data_folder}/{gadm_fn}"
             gadm = gf.from_geofeather(gadmDir)
             gadm["country"] = gadm["NAME_0"]
-            gadm["state"] = gadm["NAME_1"]
+            #gadm["state"] = gadm["NAME_1"]
             gadm["admin1"] = gadm["NAME_1"]
             gadm["admin2"] = gadm["NAME_2"]
-            gadm = gadm[["geometry", "country", "state", "admin1", "admin2"]]
+            #gadm = gadm[["geometry", "country", "state", "admin1", "admin2"]]
+            gadm = gadm[["geometry", "country", "admin1", "admin2"]]
 
         elif admin == "admin3":
             gadm_fn = f"gadm36_3.feather"
             gadmDir = f"{download_data_folder}/{gadm_fn}"
             gadm = gf.from_geofeather(gadmDir)
             gadm["country"] = gadm["NAME_0"]
-            gadm["state"] = gadm["NAME_1"]
+            #gadm["state"] = gadm["NAME_1"]
             gadm["admin1"] = gadm["NAME_1"]
             gadm["admin2"] = gadm["NAME_2"]
             gadm["admin3"] = gadm["NAME_3"]
-            gadm = gadm[["geometry", "country", "state", "admin1", "admin2", "admin3"]]        
+            #gadm = gadm[["geometry", "country", "state", "admin1", "admin2", "admin3"]]
+            gadm = gadm[["geometry", "country", "admin1", "admin2", "admin3"]]
 
     start_time = timeit.default_timer()
 
@@ -580,7 +584,11 @@ def match_geo_names(admin: str, df: pd.DataFrame, resolve_to_gadm_geotypes: list
         gadm_country_list = gadm["country"].unique()
         unknowns = df[~df.country.isin(gadm_country_list)].country.tolist()
         for unk in unknowns:
-            match = fuzzywuzzy.process.extractOne(unk, gadm_country_list, scorer=fuzz.partial_ratio)
+            try:
+                match = fuzzywuzzy.process.extractOne(unk, gadm_country_list, scorer=fuzz.partial_ratio)
+            except Exception as e:
+                match = None
+                logging.error(f"Error in match_geo_names: {e}")
             if match != None:
                 df.loc[df.country == unk, 'country'] = match[0]
 
@@ -716,6 +724,8 @@ def normalizer(df: pd.DataFrame, mapper: dict, admin: str, gadm: gpd.GeoDataFram
     df, mapper, renamed_col_dict = handle_colname_collisions(df, mapper, col_order)
 
     ### mapper is a dictionary of lists of dictionaries.
+    click.echo("Raw dataframe:")
+    click.echo(df.head())
 
     # list of names of datetime columns primary_date=True
     primary_time_cols = [k['name'] for k in mapper['date'] if 'primary_date' in k and k['primary_date'] == True]
@@ -991,7 +1001,6 @@ def normalizer(df: pd.DataFrame, mapper: dict, admin: str, gadm: gpd.GeoDataFram
                     continue
             features.append(kk)
 
-
     # Append columns annotated in feature dict to features list (if not a
     # qualifies column)
     #features.extend([k["name"] for k in mapper["feature"]])
@@ -1015,6 +1024,49 @@ def normalizer(df: pd.DataFrame, mapper: dict, admin: str, gadm: gpd.GeoDataFram
                 else:
                     qualified_col_dict[k] = [kk]
 
+        # Convert aliases based on user annotations
+        aliases = feature_dict.get("aliases", {})
+        if aliases:
+            click.echo(f"Pre-processed aliases are: {aliases}")
+            type_ = df[feature_dict["name"]].dtype.type
+            click.echo(f"Detected column type is: {type_}")
+            aliases_ = {}
+            # The goal below is to identify the data type and then to cast the 
+            # alias key from string into that type so that it will match
+            # if that fails, just cast it as a string
+            for kk, vv in aliases.items():
+                try:
+                    if issubclass(type_, (int, np.integer)):
+                        click.echo("Aliasing: integer detected")
+                        aliases_[int(kk)] = vv
+                    elif issubclass(type_, (float, np.float16, np.float32, np.float64, np.float128)):
+                        click.echo("Aliasing: float detected")
+                        aliases_[float(kk)] = vv
+                    elif issubclass(type_, (bool, np.bool, np.bool_)):
+                        click.echo("Aliasing: boolean detected")
+                        if strtobool(kk) == 1:
+                            aliases_[True] = vv
+                            click.echo("Converted true string to boolean")
+                        else:
+                            click.echo("Converted false string to boolean")
+                            aliases_[False] = vv
+                    # Fall back on string
+                    else:
+                        click.echo("Aliasing: string detected")
+                        aliases_[kk] = vv
+                except ValueError as e:
+                    # Fall back on string
+                    click.echo(f"Error: {e}")
+                    aliases_[kk] = vv
+            click.echo(f"Aliases for {feature_dict['name']} are {aliases_}.")
+            df[[feature_dict["name"]]] = df[[feature_dict["name"]]].replace(aliases_)
+            
+            # Since the user has decided to apply categorical aliases to this feature, we must coerce
+            # the entire feature to a string, even if they did not alias every value within the feature
+            # the reason for this is to avoid mixed types within the feature (e.g. half int/half string) 
+            # since this makes it difficult to visualize
+            df[[feature_dict["name"]]] = df[[feature_dict["name"]]].astype(str)
+
     # perform geocoding if lat/lng are present
     if "lat" in df and "lng" in df:
         df, df_geocode = geocode(admin, df, x="lng", y="lat", gadm=gadm, df_geocode=df_geocode)
@@ -1112,6 +1164,8 @@ def normalizer(df: pd.DataFrame, mapper: dict, admin: str, gadm: gpd.GeoDataFram
     # Handle any renamed cols being renamed.
     renamed_col_dict = audit_renamed_col_dict(renamed_col_dict)
     
+    click.echo("Processed dataframe:")
+    click.echo(df_out.head())
     return df_out[col_order], renamed_col_dict, df_geocode
 
 def optimize_df_types(df: pd.DataFrame):
@@ -1162,21 +1216,21 @@ def process(fp: str, mp: str, admin: str, output_file: str, write_output = True,
     # "meta" portion of schema specifies transformation type
     transform = mapper["meta"]
 
+    # Check transform for meta.geocode_level. Update admin to this if present.
+    if (admin == None and "geocode_level" in transform):
+        admin = transform["geocode_level"]
+
     ftype = transform["ftype"]
     if ftype == "geotiff":
-        if transform["date"] == "":
-            d = None
-        else:
-            d = transform["date"]
-
         df = raster2df(
             InRaster = fp,
             feature_name = transform["feature_name"],
-            band = int(transform["band"] if "band" in transform and transform["band"] != "" else "0"),
-            nodataval = int(transform["null_val"]),
-            date = d,
-            band_name = transform["band_name"],
-            bands = transform["bands"] if "bands" in transform else None
+            band         = int(transform["band"] if "band" in transform and transform["band"] != "" else "0"),
+            nodataval    = int(transform["null_val"]),
+            date         = transform["date"] if ("date" in transform and transform["date"] != "") else None,
+            band_name    = transform["band_name"],
+            bands        = transform["bands"] if "bands" in transform else None,
+            band_type    = transform["band_type"] if "band_type" in transform else "category"
         )
     elif ftype == 'excel':
         df = pd.read_excel(fp, transform['sheet'])
@@ -1242,7 +1296,9 @@ def raster2df(
     nodataval: int = -9999,
     date: str = None,
     band_name: str = "feature2",
-    bands: dict = None
+    bands: dict = None,
+    band_type: str = 'category'
+
 ) -> pd.DataFrame:
     """
     Description
@@ -1266,6 +1322,8 @@ def raster2df(
     bands: dict, default None
         passed in meta; dictionary of band identifiers and specifies bands to 
         be processed.
+    band_type: str, default category
+        Specifies band type e.g. category or datetime. If datetime, this data goes into the date column.
 
     Examples
     --------
@@ -1280,35 +1338,45 @@ def raster2df(
     ColRange = range(ds.RasterXSize)
     RowRange = range(ds.RasterYSize)
 
-    if band == 0 and bands:
-        # TODO Will someone import a raster but not use all bands?
-        #assert(len(bands) == ds.RasterCount)
-        pass
-
     # Cache the dataframe and value data type.
     df = pd.DataFrame()
     row_data_type = None
-
-    if bands == None:
-        columns = ["longitude", "latitude", feature_name]
-    else:
-        columns=["longitude", "latitude", feature_name, band_name]
-
+    
     for x in range(1, ds.RasterCount+1):
-        # If band has a value, then limit import to single band.
+        # If band has a value, then limit import to the single specified band.
         if band > 0 and band != x:
             continue
 
         # If no bands in meta, then single-band and use band_name
         # If bands, then process only those in the meta.
-        if bands == None:
+        if not bands:
             band_value = band_name
+            logging.info(f"Single band detected. Bands: {bands}, band_name: {band_name}, feature_name: {feature_name}")
         elif str(x) in bands:
-            band_value = bands[str(x)] 
-        else:
+            band_value = bands[str(x)]
+            logging.info(f"Multi-band detected Bands: {bands}, band_name: {band_name}, feature_name: {feature_name}")
+        elif str(x) not in bands:
+            # Processing a band not specified in the meta, so skip it
+            logging.info(f"Skipping band {x} since it is not specified in {bands}.")
             continue
+        else:
+            raise Exception(f"Neither single nor multiple bands specified in meta. Current band: {x}, Bands: {bands}, band_name: {band_name}, feature_name: {feature_name}")
+
+        # Create columns for the dataframe.
+        if not bands:
+            columns = ["longitude", "latitude", feature_name]
+            logging.info(f"Single band detected. Columns are: {columns}")
+        elif band_type == 'datetime':
+            columns=["longitude", "latitude", 'date', feature_name]
+            logging.info(f"Datetime multiband detected. Columns are: {columns}")
+        elif band_type == 'category':
+            # categorical multi-band; add columns during processing.
+            columns=["longitude", "latitude", band_value]
+            logging.info(f"Categorical multiband detected. Columns are: {columns}")
+        else:
+            raise Exception(f"During column processing, neither single nor multiple bands specified in meta. Bands: {bands}, band_name: {band_name}, feature_name: {feature_name}")
                 
-        rBand = ds.GetRasterBand(x)  # (band) # first band
+        rBand = ds.GetRasterBand(x)  
         nData = rBand.GetNoDataValue()
 
         if nData == None:
@@ -1352,17 +1420,18 @@ def raster2df(
                     # if (RowData[ThisCol] <= 100) and (RowData[ThisCol] >= -100):
 
                     X = GeoTrans[0] + (ThisCol * GeoTrans[1])
-                    Y = GeoTrans[3] + (
-                        ThisRow * GeoTrans[5]
-                    )  # Y is negative so it's a minus
+                    Y = GeoTrans[3] + (ThisRow * GeoTrans[5])  # Y is negative so it's a minus
                     # this gives the upper left of the cell, offset by half a cell to get centre
                     X += HalfX
                     Y += HalfY
 
+                    # Add the data row to the dataframe.
                     if bands == None:
                         points.append([X, Y, row_value])
+                    elif band_type == 'datetime':
+                        points.append([X, Y, band_value, row_value])
                     else:
-                        points.append([X, Y, row_value, band_value])
+                        points.append([X, Y, row_value])
 
         # This will make all floats float64, but will be optimized in process().
         new_df = pd.DataFrame(points, columns=columns)
@@ -1371,10 +1440,14 @@ def raster2df(
             df = new_df
         else:
             #df = df.merge(new_df, left_on=["longitude", "latitude"], right_on=["longitude", "latitude"])
-            df = df.append(new_df)
+            if (bands and band_type != 'datetime'):
+                #df.join(new_df, on=["longitude", "latitude"])
+                df = df.merge(new_df, left_on=["longitude", "latitude"], right_on=["longitude", "latitude"])
+            else:
+                df = df.append(new_df)
 
     # Add the date from the mapper.
-    if date:
+    if (date and band_type != 'datetime'):
         df["date"] = date
 
     df.sort_values(by=columns, inplace=True)
@@ -1391,12 +1464,15 @@ class mixdata:
         gadmDir = f"{download_data_folder}/{gadm_fn}"
         gadm = gf.from_geofeather(gadmDir)
         gadm["country"] = gadm["NAME_0"]
-        gadm["state"] = gadm["NAME_1"]
+        #gadm["state"] = gadm["NAME_1"]
         gadm["admin1"] = gadm["NAME_1"]
         gadm["admin2"] = gadm["NAME_2"]    
         gadm0 = gadm[["geometry", "country"]]
-        gadm1 = gadm[["geometry", "country", "state", "admin1"]]
-        gadm2 = gadm[["geometry", "country", "state", "admin1", "admin2"]]
+        #gadm1 = gadm[["geometry", "country", "state", "admin1"]]
+        #gadm2 = gadm[["geometry", "country", "state", "admin1", "admin2"]]
+        gadm1 = gadm[["geometry", "country", "admin1"]]
+        gadm2 = gadm[["geometry", "country", "admin1", "admin2"]]
+        
         self.gadm0 = gadm0
         self.gadm1 = gadm1
         self.gadm2 = gadm2
@@ -1409,11 +1485,12 @@ class mixdata:
         gadmDir = f"{download_data_folder}/{gadm_fn}"
         gadm3 = gf.from_geofeather(gadmDir)
         gadm3["country"] = gadm3["NAME_0"]
-        gadm3["state"] = gadm3["NAME_1"]
+        #gadm3["state"] = gadm3["NAME_1"]
         gadm3["admin1"] = gadm3["NAME_1"]
         gadm3["admin2"] = gadm3["NAME_2"]
         gadm3["admin3"] = gadm3["NAME_3"]
-        gadm3 = gadm3[["geometry", "country", "state", "admin1", "admin2", "admin3"]]
+        #gadm3 = gadm3[["geometry", "country", "state", "admin1", "admin2", "admin3"]]
+        gadm3 = gadm3[["geometry", "country", "admin1", "admin2", "admin3"]]
         self.gadm3 = gadm3
         
 
@@ -1459,9 +1536,15 @@ class mixdata:
 #geo = 'admin2'
 #outf = 'examples/causemosify-tests'
 
+#fp = "examples/causemosify-tests/example.csv"
+#mp = "examples/causemosify-tests/example.json"
+#geo = 'admin2'
+#outf = 'examples/causemosify-tests'
+
 #start_time = timeit.default_timer()
 #df = pd.DataFrame()
-#df, dct = process(fp, mp,geo, outf)
+#print('processing...')
+#df, dct = process(fp, mp, geo, outf)
 #print('process time', timeit.default_timer() - start_time)
 #cols = ['timestamp','country','admin1','admin2','admin3','lat','lng','feature','value']
 #df.sort_values(by=cols, inplace=True)
